"use strict";
this.name        = "deep_space_pirates"; 
this.author      = "eric walch"; 
this.copyright   = "� 2008-2014 Creative Commons: attribution, non-commercial, sharealike."; 
this.description = "This script places random ships in deep space."; 
this.version     = "1.9";
/*
The script does nothing when in interstellar space or when docked or when under attack.
The chance of encounters is less when traveling slow. NB jumpSpeed is about 32 * maxSpeed.
It only add ships when not within "this.spacelaneRadius" meters from the core of: the spacelane planet-witchpoint, the spacelane planet-sun or any station.
Decrease change of encounters logaritmic with distance.
*/

this.$logging = true;
this.$debug = false;
this.$spacelaneRadius = 100E3;
this.$addition = 0;
this.$preserveDerelicts = false;
this.$preserveAsteroids = false;
 // planet radius is between 2816 km and 6912 km on F7 screen, main-station at twice this distance.
 // In calculation Oolite uses internal 28160 til 69120 meters

this.$libConfig = { // for Lib_Config to handle on-station OXP configuration
    Name:   "deep_space_pirates",
    Alias:  "Deep Space Pirates",
    Display:"Config",
    Alive:  "$libConfig",
    Bool:   {
                B0: {
                        Name:   "$preserveDerelicts",
                        Desc:   "Preserve derelicts",
                        Def:    false, 
                },
                B1: {
                        Name:   "$preserveAsteroids",
                        Desc:   "Preserve asteroids",
                        Def:    false,
                },
                Info:   "false: remove when out of player's sight; true: preserve"
    }
}


//---------------------------------------------------------------------------------------------//
this.startUp = function _startUp() {
    var mv = missionVariables;
    if (!mv.deep_space_pirate_kills) mv.deep_space_pirate_kills = 1; // 0 would be a problem with Math.log()
    if (mv.deep_space_pirates_preserve_derelicts == 1) 
        this.$preserveDerelicts = true;
    else
        this.$preserveDerelicts = false;
    if (mv.deep_space_pirates_preserve_asteroids == 1)
        this.$preserveAsteroids = true;
    else
        this.$preserveAsteroids = false;
    this.$additiontime = clock.seconds;
    // With the long range scanner available, also pirates can find the player better.
    if (player.ship.equipmentStatus("EQ_LONG_RANGE_SCANNER") != "EQUIPMENT_UNKNOWN") 
        this.$addition = 10;
    this.$dsGroup = new ShipGroup("DeepSpaceShips"); // start with an empty group
}

//---------------------------------------------------------------------------------------------//
this.startUpComplete = function _startUpComplete() {
    if (worldScripts.Lib_Config) {
        this.$missionVars = missionVariables;
        var ret = worldScripts.Lib_Config._registerSet(this.$libConfig);
        if (this.$logging) log(this.name, "Lib Config registerSet returned "+ret);
    }
}

//---------------------------------------------------------------------------------------------//
this.shipLaunchedFromStation = this.shipExitedWitchspace = function _startTimer() {
    if (this.$checkTimer)
        this.$checkTimer.start()
    else
        this.$checkTimer = new Timer(this, this.$dsCheck, 30, 10)
}

//---------------------------------------------------------------------------------------------//
this.shipWillDockWithStation = this.shipWillEnterWitchspace = function _stopTimer() {
    if (this.$checkTimer && this.$checkTimer.isRunning)
        this.$checkTimer.stop();
}

//---------------------------------------------------------------------------------------------//
this.playerWillSaveGame = function _playerWillSaveGame() {
    var mv = missionVariables;
    mv.deep_space_pirates_preserve_derelicts = (this.$preserveDerelicts ? 1 : 0);
    mv.deep_space_pirates_preserve_asteroids = (this.$preserveAsteroids ? 1 : 0);
}

//---------------------------------------------------------------------------------------------//
this.$dsCheck = function _dsCheck() {
    var _ship = player.ship;
    var rnd = Math.random();
    var thresholdSpeed = 20 * _ship.maxSpeed * rnd;
    if (this.$debug) log(this.name, "timer, speed: "+_ship.speed.toFixed(0)+", threshold: "+thresholdSpeed.toFixed(0)+", speed_chk:"+(_ship.speed>thresholdSpeed)+", clock:"+(clock.seconds - this.$additiontime > 120));
    if (!system.isInterstellarSpace 
            && !_ship.docked
            && (_ship.speed > 20 * _ship.maxSpeed * rnd)
            && !player.alertHostiles 
            && (clock.seconds - this.$additiontime > 120)) 
        this.$dsAddPirate();
    
    this.$dsRemoveShips();
}

//---------------------------------------------------------------------------------------------//
this.$dsAddPirate = function _dsAddPirate() {
    var _ship = player.ship;
    var logging = this.$logging;
    var debug = this.$debug;
    var local_name = this.name;
    // first check planet-witchpoint spacelane cylinder.
    var spacelaneRadius = this.$spacelaneRadius;
    var playerPosition = _ship.position;
    var distanceToSpacelane = Math.sqrt(playerPosition.x * playerPosition.x + playerPosition.y * playerPosition.y);
    var chanceFactor = 0.2 / Math.log(distanceToSpacelane / spacelaneRadius); // value is nagative up to this.spacelaneRadius.

    // Above this.spacelaneRadius meter distance, chanceFactor starts high but decays gradually.
    if (distanceToSpacelane < 3*spacelaneRadius) chanceFactor *= 3 - system.government/3.5; // varies between 1 and 3
    if (chanceFactor < Math.random()) return; // stop further calculations.

    // calculate planet-sun spacelane cylinder.
    var sunSpacelane = new Vector3D(system.mainPlanet.position.subtract(system.sun.position));
    var sunPlayerVector = new Vector3D(playerPosition.subtract(system.sun.position));
    const distanceToSun = sunPlayerVector.magnitude();
    const distanceToSunSpacelane = distanceToSun * Math.sin(sunSpacelane.angleTo(sunPlayerVector));

    if (debug) log(this.name, "distance to land WP: "+distanceToSpacelane.toFixed(0)+", distance to lane PS: "+distanceToSunSpacelane.toFixed(0));
    if (distanceToSpacelane > spacelaneRadius && distanceToSunSpacelane > spacelaneRadius) {
        var heading = _ship.heading; 
        var forwardPosition = new Vector3D();
        var rnd = Math.random();
        if (debug) log(this.name, "random: "+rnd+", threshold: "+(0.6 - system.government / 20));
        if (rnd  <  0.6 - system.government / 20) { // 60% chance of pirate in Anarchie, 25% in Corporate State
            if (system.countShipsWithRole("station", _ship, 100E3) > 0) return;
            // no Pirate addition when > 150 or when within 100 000 m of station
            forwardPosition = playerPosition.add(heading.multiply(50E3));
            // var maxShipNumber = 10.5 + Math.log(missionVariables.deep_space_pirate_kills) - system.government/2
             // Base max is 7, in anarchies up to 3.5 ships extra and 3.2 extra for every decade of kills.
            if (this.$addition>100) this.$addition = 10; // don't let the value grow to large. 
            const maxShipNumber = Math.log(++this.$addition) + Math.log(missionVariables.deep_space_pirate_kills) / 2 + 5.5 - system.government / 2;
             // Base max is 2, in anarchies up to 3.5 ships extra and 3.2 extra for every decade of kills
             // + 3.2 extra for every decade of addition during this startup.
            var pirateGroup = system.addGroup("pirate", this.$dsBiasedRandom(maxShipNumber), forwardPosition, 5E3);
            var pirates = pirateGroup.ships;
            pirateGroup.name = "Deep Space Pirates";
            this.$dsAddShipsToQueue(pirateGroup);
            let i = pirates.length;
            while (i--) {
                if (pirates[i].escortGroup && pirates[i].escortGroup.count>1) {
                    var j = pirates[i].escortGroup.count;
                    while (j--) {
                        var escort = pirates[i].escortGroup.ships[j];
                        if (escort !== pirates[i]) {
                            escort.switchAI("DSPescortAI.plist");
                            escort.script.shipDied = this.$dsPirateDied;
                        }
                    };
                }

                if (pirates[i].cargoSpaceCapacity > 15 && pirates[i].entityPersonality&7<3) {
                    pirates[i].bounty = 0;
                    pirates[i].script.shipDied = this.$dsPirateDiedWithCargo;
                }
                else
                    pirates[i].script.shipDied = this.$dsPirateDied;
                if (pirates[i].fuel  ===  0) pirates[i].fuel = 7;
            }
            
            // Convert the script of the first ship.
            // Check if it has an old PirateAI.plist or one of the newer "oolite-piratexxxx.js" scripts. Only than change its AI behaviour
            if (pirates[0].AI  ===  "pirateAI.plist" || 
                (pirates[0].AI  === "nullAI.plist" && pirates[0].AIScript.name.indexOf("Oolite Pirate") > -1)) {
                var state = (distanceToSpacelane < 800E3) ? "SPACELANE" : "DEEPSPACE";
                if (pirates.length>10) state = "HIDEOUT";
                pirates[0].switchAI("DSPpirateAI.plist");
                pirates[0].AIState = state;
            }

            if (logging) log(local_name, "Added "+ pirates.length +" pirates out of a possible "+ Math.ceil(maxShipNumber) + ". Killcounter: " + missionVariables.deep_space_pirate_kills);
            this.$additiontime = clock.seconds;
        } else {
            if (system.countShipsWithRole("station", _ship, 250E3) > 0) return;
            // no when within 250 000 m of station
            var ship;
            var chance = Math.random();
            var count = 1;
            
            if (chance < 0.10)
                ship = "trader";
            else if (chance < 0.20)
                ship = "trader-courier";
            else if (chance < 0.30)
                ship = "trader-smuggler";
            else if (chance < 0.40)
                ship = "hunter";
            else if (chance < 0.50)
                ship = system.techLevel < 8 ? "police" : "interceptor";
            else if (chance < 0.99) {
                ship = "asteroid";
                count = this.$dsBiasedRandom(6);
            } else
                ship = "RockHermit";
            
            if (logging) log(local_name, "Adding " + count + " " + ship);
            if (ship != "RockHermit") {
                forwardPosition = playerPosition.add(heading.multiply(90E3));
                this.$dsAddShipsToQueue(system.addGroup(ship, count, forwardPosition, 10E3));
            } else {
                forwardPosition = playerPosition.add(heading.multiply(200E3));
                this.$dsAddShipsToQueue(system.addGroup("DSProckhermit", 1, forwardPosition, 10E3));
                this.$dsAddShipsToQueue(system.addGroup("asteroid", 8, forwardPosition, 10E3));
                // The only reason for using a custom RockHermit is that we had a yellow scansclass in the past
                // when adding the build-in one. With Oolite 1.80 this is probably no longer needed.
            }
        }
    } else if (debug)
        log(this.name, "Too close to the lanes: "+distanceToSpacelane.toFixed(0)+" to PW, "+distanceToSunSpacelane.toFixed(0)+" to PS");
}

//---------------------------------------------------------------------------------------------//
this.$dsAddShipsToQueue = function _dsAddShipsToQueue(group) {
    if (group && group.count > 0) {
        var i = group.count;
        while (i--) {
            this.$dsGroup.addShip(group.ships[i]);
            if (group.ships[i].escortGroup.count > 1) {
                var j = group.ships[i].escortGroup.count;
                while (j--)
                    this.$dsGroup.addShip(group.ships[i].escortGroup.ships[j]);
            }
        }
    }
    if (this.$logging) log(this.name, "Shipgroup: " + this.$dsGroup.ships);
}

//---------------------------------------------------------------------------------------------//
this.$dsRemoveShips = function _dsRemoveShips() {
    var dsGroup = this.$dsGroup;
    var local_name = this.name;
    var logging = this.$logging;
    var mv = missionVariables;
    var removeDerelicts = !this.$preserveDerelicts;
    var removeAsteroids = !this.$preserveAsteroids;
    var i = dsGroup.count;

    while (i--) {
        var s = dsGroup.ships[i];
        if (!s.isVisible && (removeDerelicts || !s.isDerelict) && (removeAsteroids || !s.isBoulder)) {
            if (logging) log(local_name, "removing: " + s.primaryRole+" ("+s.displayName+")");
            s.remove(true);
        }
    }
}

//---------------------------------------------------------------------------------------------//
this.$dsBiasedRandom = function _dsBiasedRandom(amount) {
    return Math.ceil(Math.pow(Math.random(), 3) * amount);
}

// next code is used by ship scripts.
//---------------------------------------------------------------------------------------------//
this.$dsPirateDied = function _dsPirateDied(whom) {
    if (whom && whom.isPlayer && !player.ship.isCloaked) missionVariables.deep_space_pirate_kills++;
}

//---------------------------------------------------------------------------------------------//
this.$dsPirateDiedWithCargo = function _dsPirateDiedWithCargo(whom) {
    if (whom && whom.isPlayer && !player.ship.isCloaked) missionVariables.deep_space_pirate_kills++;
    var cargo = Math.ceil(10 * Math.random());
    if (this.ship.cargoSpaceAvailable < 8) cargo = this.ship.cargoSpaceAvailable;
    if (cargo > 0) this.ship.spawn("cargopod", cargo);
}

//---------------------------------------------------------------------------------------------//
this.$dsShipAcceptedEscort = function _dsShipAcceptedEscort(escort) {
        if (escort.script.asigned) return; // escort is already initialised by another ds mother that now has died.
        if (escort.bounty === 0) escort.bounty = this.ship.bounty + 5;
        escort.switchAI("DSPescortAI.plist");
        escort.script.shipDied = worldScripts["deep_space_pirates"].dsPirateDied;
        escort.script.asigned = true;
        if (this.$logging) log(this.name, this.ship.displayName+" accepted escort");
}
